#!/usr/bin/env python3

## Generic model for Pytorch NN
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Classes :
# - ModelPyTorch -> Generic model for Pytorch NN


import os
import time
import dill
import json
import ntpath
import shutil
import logging
import functools
import numpy as np
import pandas as pd
from typing import Union, Any, List

import torch
import pytorch_lightning as pl
from torch.utils.data import DataLoader
from pytorch_lightning import loggers as pl_loggers
from pytorch_lightning.callbacks import ModelCheckpoint, EarlyStopping, LearningRateMonitor, ProgressBar

from {{package_name}} import utils
from {{package_name}}.models_training.model_class import ModelClass
from {{package_name}}.models_training import utils_deep_torch, utils_models

import matplotlib.pyplot as plt
import seaborn as sns

sns.set(style="darkgrid")


class ModelPyTorch(ModelClass):
    '''Generic model for Pytorch NN'''

    _default_name = 'model_pytorch'

    # Not implemented :
    # -> predict_proba
    # -> _get_train_dataloader
    # -> _prepare_x_test
    # -> _get_model
    # -> reload

    def __init__(self, batch_size: int = 64, epochs: int = 99, validation_split: float = 0.2, patience: int = 5,
                 embedding_name: str = 'cc.fr.300.pkl', pytorch_params: Union[dict, None] = None, **kwargs) -> None:
        '''Initialization of the class (see ModelClass for more arguments)

        Kwargs:
            batch_size (int): Batch size
            epochs (int): Number of epochs
            validation_split (float): Percentage for the validation set split
                Only used if no input validation set when fitting
            patience (int): Early stopping patience
            embedding_name (str) : The name of the embedding matrix to use
            pytorch_params (dict): Parameters used by pytorch models
                e.g. learning_rate, nb_lstm_units, etc...
                The purpose of this dictionary is for the user to use it as they wants in the _get_model function
                This parameter was initially added in order to do an hyperparameters search
        '''
        # TODO: learning rate should be an attribute !
        # Init.
        super().__init__(**kwargs)

        # Get logger (must be done after super init)
        self.logger = logging.getLogger(__name__)

        # Param. model
        self.batch_size = batch_size
        self.epochs = epochs
        self.validation_split = validation_split
        self.patience = patience

        # Model set on fit
        self.model: Any = None

        # Param embedding (can be None if no embedding)
        self.embedding_name = embedding_name

        # Pytorch params
        if pytorch_params is None:
            pytorch_params = {}
        self.pytorch_params = pytorch_params.copy()

    def fit(self, x_train, y_train, x_valid=None, y_valid=None, with_shuffle: bool = True, **kwargs) -> None:
        '''Fits the model

        Args:
            x_train (?): Array-like, shape = [n_samples, n_features]
            y_train (?): Array-like, shape = [n_samples, n_targets]
            x_valid (?): Array-like, shape = [n_samples, n_features]
            y_valid (?): Array-like, shape = [n_samples, n_targets]
        Kwargs:
            with_shuffle (bool): If x, y must be shuffled before fitting
                Experimental: We must verify if it works as intended depending on the formats of x and y
                This should be used if y is not shuffled as the split_validation takes the lines in order.
                Thus, the validation set might get classes which are not in the train set ...
        Raises:
            AssertionError: If different classes when comparing an already fitted model and a new dataset
        '''
        ##############################################
        # Manage retrain
        ##############################################

        # If a model has already been fitted, we make a new folder in order not to overwrite the existing one !
        # And we save the old conf
        # TODO: Do everything again !!!
        if self.trained:
            # Get src files to save
            src_files = [os.path.join(self.model_dir, "configurations.json")]
            if self.nb_fit > 1:
                for i in range(1, self.nb_fit):
                    src_files.append(os.path.join(self.model_dir, f"configurations_fit_{i}.json"))
            # Change model dir
            self.model_dir = self._get_model_dir()
            # Get dst files
            dst_files = [os.path.join(self.model_dir, f"configurations_fit_{self.nb_fit}.json")]
            if self.nb_fit > 1:
                for i in range(1, self.nb_fit):
                    dst_files.append(os.path.join(self.model_dir, f"configurations_fit_{i}.json"))
            # Copies
            for src, dst in zip(src_files, dst_files):
                try:
                    shutil.copyfile(src, dst)
                except Exception as e:
                    self.logger.error(f"Impossible to copy {src} to {dst}")
                    self.logger.error("We still continue ...")
                    self.logger.error(repr(e))

        ##############################################
        # Prepare x_train, x_valid, y_train & y_valid
        # Also extract list of classes
        ##############################################

        # If not multi-labels, transform y_train as dummies (should already be the case for multi-labels)
        if not self.multi_label:
            # If len(array.shape)==2, we flatten the array if the second dimension is useless
            if isinstance(y_train, np.ndarray) and len(y_train.shape) == 2 and y_train.shape[1] == 1:
                y_train = np.ravel(y_train)
            if isinstance(y_valid, np.ndarray) and len(y_valid.shape) == 2 and y_valid.shape[1] == 1:
                y_valid = np.ravel(y_valid)
            # Transformation dummies
            y_train_dummies = pd.get_dummies(y_train)
            y_valid_dummies = pd.get_dummies(y_valid) if y_valid is not None else None
            # Important : get_dummies reorder the columns in alphabetical order
            # Thus, there is no problem if we fit again on a new dataframe with shuffled data
            list_classes = list(y_train_dummies.columns)
        # Else keep it as it
        else:
            y_train_dummies = y_train
            y_valid_dummies = y_valid
            if hasattr(y_train_dummies, 'columns'):
                list_classes = list(y_train_dummies.columns)
            else:
                self.logger.warning(
                    "Can't read the name of the columns of y_train -> inverse transformation won't be possible"
                )
                # We still create a list of classes in order to be compatible with other functions
                list_classes = [str(_) for _ in range(pd.DataFrame(y_train_dummies).shape[1])]

        # Set dict_classes based on list classes
        dict_classes = {i: col for i, col in enumerate(list_classes)}

        # Validate classes if already trained, else set them
        if self.trained:
            assert self.list_classes == list_classes, \
                "Error: the new dataset does not match with the already fitted model"
            assert self.dict_classes == dict_classes, \
                "Error: the new dataset does not match with the already fitted model"
        else:
            self.list_classes = list_classes
            self.dict_classes = dict_classes

        # Shuffle x, y if wanted (advised)
        # Hence we might have classes in the validation data that we never met in the training data
        if with_shuffle:
            p = np.random.permutation(len(x_train))
            x_train = np.array(x_train)[p]
            y_train_dummies = np.array(y_train_dummies)[p]
        # Else still transform to numpy array
        else:
            x_train = np.array(x_train)
            y_train_dummies = np.array(y_train_dummies)

        # Also get y_valid_dummies as numpy
        y_valid_dummies = np.array(y_valid_dummies)

        ##############################################
        # Apply validation_split if no validation given
        ##############################################

        if y_valid is None:
            self.logger.info(f"No validation set, we split the train set : ({round((1 - self.validation_split) * 100, 2)} % train, {round((self.validation_split) * 100, 2)} % validation)")
            p = np.random.permutation(int(len(x_train) * (1 - self.validation_split)))
            mask = np.ones(len(x_train), np.bool)
            mask[p] = 0
            x_valid = x_train[mask]
            y_valid_dummies = y_train_dummies[mask]
            x_train = x_train[~mask]
            y_train_dummies = y_train_dummies[~mask]

        ##############################################
        # Get data loaders
        ##############################################

        # Get train dataloader
        self.logger.info("Preparing training set")
        train_dl = self._get_train_dataloader(self.batch_size, x_train, y_train_dummies)

        # Also get valid data loader
        self.logger.info("Preparing validation set")
        valid_dl = self._get_test_dataloader(self.batch_size, x_valid, y_valid_dummies)  # We could change the batch size value

        ##############################################
        # Fit
        ##############################################

        self.logger.info(f"Training ;..")

        # Get model (if already fitted we do not load a new one)
        if not self.trained:
            self.model = self._get_model(len(train_dl))

        # Get callbacks (early stopping & checkpoint)
        callbacks = self._get_callbacks()

        # Get logger
        # logger = self._get_logger()
        # TODO: Understand why the training won't work when the logger is added
        logger = False

        # Fit
        # We use a try...except in order to save the model if an error arises
        # after more than a minute into training
        start_time = time.time()
        try:
            # We unfreeze the weights
            self.model.unfreeze()  # type: ignore
            # We go to train mode (it should be already the case, but one never knows !)
            self.model.train()  # type: ignore
            # We train
            trainer = pl.Trainer(
                default_root_dir=self.model_dir,
                max_epochs=self.epochs,
                gradient_clip_val=self.model.gradient_clip_val if hasattr(self.model, 'gradient_clip_val') else None,  # type: ignore
                weights_summary="top",  # "full" would print all layers summary
                gpus=1 if torch.cuda.is_available() else 0,
                callbacks=callbacks,
                logger=logger,
            )
            trainer.fit(self.model, train_dl, valid_dl)  # type: ignore
            # If checkpoint best_model, we reload
            if os.path.exists(os.path.join(self.model_dir, 'best_model.ckpt')):
                self.logger.info("Loading of the best model")
                self.model = self.reload_model(os.path.join(self.model_dir, 'best_model.ckpt'))
            else:
                self.logger.warning("Can't load the best model !")
            # And we freeze the weights again
            self.model.freeze()  # type: ignore
        except (RuntimeError, SystemError, SystemExit, EnvironmentError, KeyboardInterrupt, Exception) as e:
            # Steps:
                # 1. Display error
                # 2. Check if more than one minute elapsed & not several iterations & existence best.hdf5
                # 3. Reload best model
                # 4. We consider that a fit occured (trained = True, nb_fit += 1)
                # 5. Save & create a warning file
                # 6. Display error messages
                # 7. Raise an error

            # 1.
            self.logger.error(repr(e))
            raise RuntimeError("Error during model training")
            # TODO : do as in keras -> save best model and reload

            # # Print accuracy & loss if level_save > 'LOW'
            # TODO : do as in keras -> plot metrics
            # if self.level_save in ['MEDIUM', 'HIGH']:
            #     self._plot_metrics_and_loss(fit_history, iter)
            #     # Reload best model
            #     self.model = load_model(
            #         os.path.join(self.model_dir, f'best.hdf5'),
            #         custom_objects=self.custom_objects
            #     )

        # Set trained
        self.trained = True
        self.nb_fit += 1

    @utils.data_agnostic_str_to_list
    @utils.trained_needed
    def predict(self, x_test, return_proba: bool = False, **kwargs) -> np.ndarray:
        '''Predictions on test set

        Args:
            x_test (?): Array-like or sparse matrix, shape = [n_samples]
        Kwargs:
            return_proba (bool): If the function should return the probabilities instead of the classes
        Returns:
            (np.ndarray): Array, shape = [n_samples, n_classes]
        '''
        # Cast en pd.Series
        x_test = pd.Series(x_test)

        # Get predictions per chunk
        predicted_proba = self.predict_proba(x_test)

        # We return the probabilities if wanted
        if return_proba:
            return predicted_proba

        # Finally, we get the classes predictions
        return self.get_classes_from_proba(predicted_proba)

    @utils.data_agnostic_str_to_list
    @utils.trained_needed
    def predict_proba(self, x_test, **kwargs) -> np.ndarray:
        '''Probabilities predicted on the test set

        Args:
            x_test (?): Array-like or sparse matrix, shape = [n_samples]
        Returns:
            (np.ndarray): Array, shape = [n_samples, n_classes]
        '''
        raise NotImplementedError("'predict_proba' needs to be overridden")

    def _get_train_dataloader(self, batch_size: int, x_train, y_train_dummies = None) -> DataLoader:
        '''Prepares the input data for the model

        Args:
            batch_size (int): Train batch size
            x_train (?): Array-like, shape = [n_samples, n_features]
            y_train_dummies (?): Array-like, shape = [n_samples, n_features]
        Returns:
            (?): Data loader
        '''
        raise NotImplementedError("'_get_train_dataloader' needs to be overridden")

    def _get_test_dataloader(self, batch_size: int, x_test, y_test_dummies=None) -> DataLoader:
        '''Prepares the input data for the model

        Args:
            x_test (?): Array-like, shape = [n_samples, n_features]
            batch_size (int): Test batch size
        Kwargs:
            y_test_dummies (?): Array-like, shape = [n_samples, n_features]
        Returns:
            (?): Data loader
        '''
        raise NotImplementedError("'_get_test_dataloader' needs to be overridden")

    # TODO: redo with tokenizer pytorch
    # def _get_embedding_matrix(self, tokenizer):
    #     '''Get embedding matrix
    #
    #     Args:
    #         tokenizer (?): Tokenizer to use (useful to test with a new matrice embedding)
    #     Returns:
    #         np.ndarray: embedding matrix
    #         int: embedding size
    #     '''
    #     # Get embedding indexes
    #     embedding_indexes = utils_models.get_embedding(self.embedding_name)
    #     # Get embedding_size
    #     embedding_size = len(embedding_indexes[list(embedding_indexes.keys())[0]])
    #
    #     # Get embedding matrix
    #     # The first line of this matrix is a zero vector
    #     # The following lines are the projections of the words obtained by the tokenizer (same index)
    #
    #     # We keep only the max tokens 'num_words'
    #     # https://github.com/keras-team/keras/issues/8092
    #     if tokenizer.num_words is None:
    #         word_index = {e: i for e, i in tokenizer.word_index.items()}
    #     else:
    #         word_index = {e: i for e, i in tokenizer.word_index.items() if i <= tokenizer.num_words}
    #     # Create embedding matrix
    #     embedding_matrix = np.zeros((len(word_index) + 1, embedding_size))
    #     # Fill it
    #     for word, i in word_index.items():
    #         embedding_vector = embedding_indexes.get(word)
    #         if embedding_vector is not None:
    #             # words not found in embedding index will be all-zeros.
    #             embedding_matrix[i] = embedding_vector
    #     self.logger.info(f"Size of the embedding matrix (ie. number of matches on the input) : {len(embedding_matrix)}")
    #     return embedding_matrix, embedding_size

    def _get_model(self, train_dataloader_size: int = None) -> Any:
        '''Gets a model structure

        Kwargs:
            train_dataloader_size (int): number of batch per epochs. Useful to set a learning rate scheduler
        Returns:
            (?): a PyTorch model
        '''
        raise NotImplementedError("'_get_model' needs to be overridden")

    def _get_callbacks(self, **kwargs) -> list:
        '''Gets model callbacks

        Returns:
            list: List of callbacks
        '''
        callbacks: List[Any] = []
        # Early stopping
        if self.patience > 1:
            self.logger.info(f"Add Early Stopping (patience = {self.patience})")
            es = EarlyStopping(monitor='val_loss', patience=self.patience)
            callbacks.append(es)
        # Add checpoint callback
        if self.level_save in ['MEDIUM', 'HIGH']:
            checkpoint_callback = ModelCheckpoint(monitor='val_loss',
                                                  save_top_k=1,
                                                  dirpath=self.model_dir,
                                                  save_weights_only=True,
                                                  filename='best_model')
            callbacks.append(checkpoint_callback)
        # Add a progress bar that does not overwrite previous epoch
        pb = MyProgressBar()
        callbacks.append(pb)
        return callbacks

    def _get_logger(self) -> Any:
        '''Gets model logger

        Returns:
            ?: Logger
        '''
        # TODO: The logger crashes the model, thus we do not use it ...

        # # Def. tensorboard
        # models_path = utils.get_models_path()
        # tensorboard_dir = os.path.join(models_path, 'tensorboard_logs')
        # # We add a prefix so that the function load_model works correctly (it looks for a sub-folder with model name)
        # log_dir = os.path.join(tensorboard_dir, f"tensorboard_{ntpath.basename(self.model_dir)}")
        # if not os.path.exists(log_dir):
        #     os.makedirs(log_dir)
        # logger = pl_loggers.TensorBoardLogger(log_dir)

        # CSV logger
        csv_dir = os.path.join(self.model_dir, 'csv_logs')
        if not os.path.exists(csv_dir):
            os.makedirs(csv_dir)
        logger = pl_loggers.CSVLogger(csv_dir)
        return logger

    # TODO plot metrics
    # def _plot_metrics_and_loss(self, fit_history) -> None:
    #     '''Function to plot accuracy & loss
    #
    #     Arguments:
    #         fit_history (?) : fit history
    #     '''
    #     # Manage dir
    #     plots_path = os.path.join(self.model_dir, 'plots')
    #     if not os.path.exists(plots_path):
    #         os.makedirs(plots_path)
    #
    #     # Get a dictionnary of possible metrics/loss plots
    #     metrics_dir = {
    #         'acc': ['Accuracy', 'accuracy'],
    #         'loss': ['Loss', 'loss'],
    #         'categorical_accuracy': ['Categorical accuracy', 'categorical_accuracy'],
    #         'f1': ['F1-score', 'f1_score'],
    #         'precision': ['Precision', 'precision'],
    #         'recall': ['Recall', 'recall'],
    #     }
    #
    #     # Plot each available metric
    #     for metric in fit_history.history.keys():
    #         if metric in metrics_dir.keys():
    #             title = metrics_dir[metric][0]
    #             filename = metrics_dir[metric][1]
    #             plt.figure(figsize=(10, 8))
    #             plt.plot(fit_history.history[metric])
    #             plt.plot(fit_history.history[f'val_{metric}'])
    #             plt.title(f"Model {title}")
    #             plt.ylabel(title)
    #             plt.xlabel('Epoch')
    #             plt.legend(['Train', 'Validation'], loc='upper left')
    #             # Save
    #             filename == f"{filename}.jpeg" if iter == 0 else f"{filename}_{iter}.jpeg"
    #             plt.savefig(os.path.join(plots_path, filename))
    #
    #             # Close figures
    #             plt.close('all')

    def save(self, json_data: Union[dict, None] = None) -> None:
        '''Saves the model

        Kwargs:
            json_data (dict): Additional configurations to be saved
        '''
        # Save configuration JSON
        if json_data is None:
            json_data = {}

        json_data['librairie'] = 'pytorch'
        json_data['batch_size'] = self.batch_size
        json_data['epochs'] = self.epochs
        json_data['validation_split'] = self.validation_split
        json_data['patience'] = self.patience
        json_data['embedding_name'] = self.embedding_name
        json_data['pytorch_params'] = self.pytorch_params
        if self.model is not None:
            json_data['pytorch_model'] = str(self.model)
        else:
            json_data['pytorch_model'] = None

        # Add _get_model code if not in json_data
        if '_get_model' not in json_data.keys():
            json_data['_get_model'] = dill.source.getsourcelines(self._get_model)[0]

        # Save strategy :
        # - best_model already saved in fit()
        # - can't pickle pytorch model, so we drop it, save, and reload it
        pytorch_model = self.model
        self.model = None
        super().save(json_data=json_data)
        self.model = pytorch_model

    def reload_model(self, model_path: str, **kwargs) -> Any:
        '''Reloads a model saved with ModelCheckpoint

        Args:
            model_path (str): model path (either checkpoint path or full torch dir, depends on subclass)
        Kwargs:
            kwargs: Dict of kwargs to override predefined params (TO BE CHECKED !!!)
        '''
        raise NotImplementedError("'reload_model' needs to be overridden")

    def reload_from_standalone(self, **kwargs) -> None:
        '''Reloads a model from its configuration and "standalones" files
        - /!\\ Needs to be overridden /!\\ -
        '''
        raise NotImplementedError("'reload' needs to be overridden")

    def _is_gpu_activated(self) -> bool:
        '''Checks if a GPU is used

        Returns:
            bool: whether GPU is available or not
        '''
        # Check for available GPU devices
        if torch.cuda.is_available():
            return True
        else:
            return False

class MyProgressBar(ProgressBar):
    '''Progress bar that does not override epochs'''
    def on_epoch_start(self, trainer, pl_module) -> None:
        print('\n\n')
        super().on_epoch_start(trainer, pl_module)


if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3
# type: ignore

## Utils - tools-functions for Faster RCNN model
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#


import logging
from typing import Callable

import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer
from tensorflow.keras.metrics import categorical_crossentropy

from {{package_name}} import utils

# Get logger
logger = logging.getLogger(__name__)


class RoiPoolingLayer(Layer):
    '''Layer selecting a zone from a ROI in a features map and resize it

    # Input shape
        List of two 4D tensors [X_img, X_roi] with shape:
        X_img : list of images
            (batch_size, cols, rows, channels)
        X_roi : list of ROI with 4 coordinates (x, y, w, h)
            (batch_size, nb_rois, 4)
    # Output shape
        5D tensor with shape:
            (batch_size, nb_rois, pool_size, pool_size, channels)
            pool_size is a parameter of resizing of the features map
    '''
    def __init__(self, pool_size: int, **kwargs) -> None:
        '''Initialization of the layer

        Args:
            pool_size (int): Output size of the layer
        '''
        self.pool_size = pool_size
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.nb_channels = input_shape[0][3]

    def compute_output_shape(self, input_shape):
        return None, None, self.pool_size, self.pool_size, self.nb_channels

    def cut_feature_map(self, feature_map, roi):
        '''Cuts a features map with a ROI

        Args:
            feature_map : input features map
                # Shape : (cols, rows, channels)
            roi : input ROI
                # Shape : (4,)
        '''
        x, y, h, w = roi[0], roi[1], roi[2], roi[3]
        return tf.image.resize(feature_map[y:y+h, x:x+w, :], (self.pool_size, self.pool_size))

    def call(self, x: list, mask=None):
        '''Call to the layer

        Args:
            x (list): List of two tensors
                0 -> features maps # Shape (batch_size, cols, rows, channels)
                1 -> rois # Shape (batch_size, nb_rois, 4)
        Returns:
            tensor: images (features maps) cut with the ROIs
                # Shape (batch_size, nb_rois, cols, rows, nb_channel)
        '''
        # Get the tensors
        feature_maps = x[0]  # Shape (batch_size, cols, rows, channels)
        rois = K.cast(x[1], 'int32')  # Shape (batch_size, nb_rois, 4)

        # We loop on each batch, and then, we loop on each ROI
        # We crop each image with the ROIs of the batch
        # We also resize (cf. pool_size)
        # We format the results and return it
        # TODO: WARNING, WE MUST HAVE THE SAME NUMBER OF ROIS PER IMAGE :'(
        output = tf.map_fn(lambda batch:
            # IN : fmap (h, w, nb_channel)
            # IN : rois (nb_rois, 4)
            tf.map_fn(
                # IN : fmap (h, w, nb_channel)
                # IN : roi (4,)
                lambda roi: self.cut_feature_map(batch[0], roi)
                , batch[1], fn_output_signature=tf.float32
                # OUT (pool_size, pool_size, nb_channel)
            )
            # OUT : (nb_rois, pool_size, pool_size, nb_channel)
        , (feature_maps, rois), fn_output_signature=tf.float32)
        # OUT : (batch_size, nb_rois, pool_size, pool_size, nb_channel)
        return output

    def get_config(self) -> dict:
        config = {'pool_size': self.pool_size}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


def get_rpn_loss_regr(nb_anchors: int) -> Callable:
    '''Gets the RPN regression loss depending on the number of anchor of the model

    Args:
        nb_anchors (int): Number of anchors of the model
    Returns:
        Callable: RPN regression loss
    '''
    def rpn_loss_regr(y_true, y_pred) -> float:
        '''Calculates the RPN regression loss (Huber loss)
               0.5*x² (if x_abs < 1)
               x_abs - 0.5 (otherwise)

        Args:
            y_true: Model's targets. Shape (batch_size, height, width, 2*4*nb_anchors)
                first part : class of the anchor boxes -> object or background
                                the loss is calculated only on anchor boxes associated to an object
                second part : regression
            y_pred: Outputs of the model. Shape (batch_size, height, width, 4*nb_anchors)
        Returns:
            float: Calculated loss
        '''
        y_true = tf.cast(y_true, tf.float32)
        y_pred = tf.cast(y_pred, tf.float32)

        # Evaluate difference
        ind_sep = 4 * nb_anchors  # Separation index of the two parts of y_true
        x = y_true[:, :, :, ind_sep:] - y_pred
        x_abs = K.abs(x)

        # If x_abs <= 1.0, x_bool = 1
        x_bool = K.cast(K.less_equal(x_abs, 1.0), tf.float32)

        # robust loss function (smooth L1)
        return K.sum(y_true[:, :, :, :ind_sep] * (x_bool * (0.5 * x * x) + (1 - x_bool) * (x_abs - 0.5))) / K.sum(1e-4 + y_true[:, :, :, :ind_sep])
    # Return loss
    return rpn_loss_regr


def get_rpn_loss_cls(nb_anchors: int) -> Callable:
    '''Gets the RPN classification loss depending on the number of anchor of the model

    Args:
        nb_anchors (int): Number of anchors of the model
    Returns:
        Callable: RPN classification loss
    '''
    def rpn_loss_cls(y_true, y_pred) -> float:
        '''Calculates the RPN classification loss (Cross entropy)

        Args:
            y_true: Model's targets. Shape (batch_size, height, width, 2*nb_anchors)
                first part : validity of the anchor box. Valid if positive (object match) or negative (background match),
                                                         Not valid if neutral (in between object and background)
                             the loss is calculated only on valid anchor boxes
                second part : classe of the anchor box
                                 0  --> background
                                 1  --> object
            y_pred: Outputs of the model. Shape (batch_size, height, width, nb_anchors)
        Returns:
            float: Calculated loss
        '''
        ind_sep = nb_anchors  # Separation index of the two parts of y_true
        return K.sum(y_true[:, :, :, :ind_sep] * K.binary_crossentropy(y_pred[:, :, :, :], y_true[:, :, :, ind_sep:])) / K.sum(1e-4 + y_true[:, :, :, :ind_sep])
    # Return loss
    return rpn_loss_cls


def get_class_loss_regr(nb_classes: int) -> Callable:
    '''Gets the classifier regression loss depending on the number of classes of the model

    Args:
        nb_classes (int): Number of classes of the model
    Returns:
        Callable: Classifier regression loss
    '''
    def class_loss_regr(y_true, y_pred):
        '''Calculates the classifier regression loss (Huber loss)
               0.5*x² (if x_abs < 1)
               x_abs - 0.5 (otherwise)

        Args:
            y_true: Model's targets. Shape (batch_size, nb_bboxes, 2*4*num_classes)
                first part : ROI class
                                  Example with two classes (without taking 'bg' into account)
                                    -> [0, 0, 0, 0, 0, 0, 0, 0] : background
                                    -> [0, 0, 0, 0, 1, 1, 1, 1] : classe n°1
                                    etc...
                                    the loss is calculated only on ROIs associated to an object
                second part : regression
            y_pred: Outputs of the model. Shape (batch_size, nb_bboxes, 4*num_classes)
        Returns:
            float: Calculated loss
        '''
        y_true = tf.cast(y_true, tf.float32)
        y_pred = tf.cast(y_pred, tf.float32)

        # Evaluate difference
        ind_sep = 4 * nb_classes  # Separation index of the two parts of y_true
        x = y_true[:, :, ind_sep:] - y_pred
        x_abs = K.abs(x)

        # If x_abs <= 1.0, x_bool = 1
        x_bool = K.cast(K.less_equal(x_abs, 1.0), tf.float32)

        return K.sum(y_true[:, :, :ind_sep] * (x_bool * (0.5 * x * x) + (1 - x_bool) * (x_abs - 0.5))) / K.sum(1e-4 + y_true[:, :, :ind_sep])
    # Return loss
    return class_loss_regr


def class_loss_cls(y_true, y_pred):
    '''Calculates the classifier classification loss (Cross entropy)

    Args:
        y_true: Model's target. Shape (batch_size, nb_bboxes, nb_classes)
        y_pred: Outputs of the model. Shape (batch_size, nb_bboxes, nb_classes)
    Returns:
        float: Calculated loss
    '''
    return K.mean(categorical_crossentropy(y_true, y_pred))


def get_custom_objects_faster_rcnn(nb_anchors: int, nb_classes: int) -> dict:
    '''Gets the keras custom_objects depending of the number of anchors and of classes of a model

    Args:
        nb_anchors (int): Number of anchors of the model
        nb_classes (int): Number of classes of the model
    Returns:
        dict: Set of customs objects
    '''

    # /!\ Important -> This dictionary defines the "custom" objets used in our Faster RCNN models
    # /!\ Important -> They are mandatory in order to serialize and save the model
    # /!\ Important -> All customs objects must be added to it
    custom_objects_faster_rcnn = {
        'RoiPoolingLayer': RoiPoolingLayer,
        'rpn_loss_regr': get_rpn_loss_regr(nb_anchors),
        'rpn_loss_cls': get_rpn_loss_cls(nb_anchors),
        'class_loss_cls': class_loss_cls,
        'class_loss_regr': get_class_loss_regr(nb_classes),
    }
    return custom_objects_faster_rcnn


if __name__ == '__main__':
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import json
import shutil
import numpy as np
import pandas as pd

from sklearn.svm import LinearSVC
from sklearn.pipeline import Pipeline
from sklearn.multiclass import OneVsRestClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_extraction.text import TfidfVectorizer

from {{package_name}} import utils
from {{package_name}}.models_training.model_pipeline import ModelPipeline

# Disable logging
import logging
logging.disable(logging.CRITICAL)


def remove_dir(path):
    if os.path.isdir(path): shutil.rmtree(path)


class ModelPipelineTests(unittest.TestCase):
    '''Main class to test model_pipeline'''

    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

    def test01_model_pipeline_init(self):
        '''Test of the initialization of {{package_name}}.models_training.model_pipeline.ModelPipeline'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Init., test all parameters
        model = ModelPipeline(model_dir=model_dir)
        self.assertEqual(model.model_dir, model_dir)
        self.assertTrue(os.path.isdir(model_dir))
        self.assertEqual(model.pipeline, None)
        # We test display_if_gpu_activated and _is_gpu_activated just by calling them
        self.assertTrue(type(model._is_gpu_activated()) == bool)
        model.display_if_gpu_activated()
        remove_dir(model_dir)

        tfidf = TfidfVectorizer()
        svc = LinearSVC()
        pipeline = Pipeline([('tfidf', tfidf), ('svc', svc)])
        model = ModelPipeline(model_dir=model_dir, pipeline=pipeline)
        model.save()
        self.assertEqual(model.pipeline, pipeline)
        remove_dir(model_dir)

    def test02_model_pipeline_fit(self):
        '''Test of the method fit of {{package_name}}.models_training.model_pipeline.ModelPipeline'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        x_train = np.array(["ceci est un test", "pas cela", "cela non plus", "ici test", "là, rien!"])
        y_train_mono = np.array([0, 1, 0, 1, 0])
        y_train_multi = pd.DataFrame({'test1': [0, 0, 0, 1, 0], 'test2': [1, 0, 0, 0, 0], 'test3': [0, 0, 0, 1, 0]})
        cols = ['test1', 'test2', 'test3']

        # Mono-label
        tfidf = TfidfVectorizer()
        svc = LinearSVC()
        pipeline = Pipeline([('tfidf', tfidf), ('svc', svc)])
        model = ModelPipeline(model_dir=model_dir, pipeline=pipeline, multi_label=False)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        self.assertFalse(hasattr(model.pipeline['svc'], "classes_"))
        model.fit(x_train, y_train_mono)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertTrue(hasattr(model.pipeline['svc'], "classes_"))
        remove_dir(model_dir)

        # We also test that the presence of arguments such as x_valid has no impact
        tfidf = TfidfVectorizer()
        svc = LinearSVC()
        pipeline = Pipeline([('tfidf', tfidf), ('svc', svc)])
        model = ModelPipeline(model_dir=model_dir, pipeline=pipeline, multi_label=False)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        self.assertFalse(hasattr(model.pipeline['svc'], "classes_"))
        model.fit(x_train, y_train_mono, x_valid='toto', y_valid='titi', test=5)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertTrue(hasattr(model.pipeline['svc'], "classes_"))
        remove_dir(model_dir)

        # Multi-labels
        tfidf = TfidfVectorizer()
        svc = LinearSVC()
        pipeline = Pipeline([('tfidf', tfidf), ('svc', OneVsRestClassifier(svc))])
        model = ModelPipeline(model_dir=model_dir, pipeline=pipeline, multi_label=True)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        self.assertFalse(hasattr(model.pipeline, "classes_"))
        model.fit(x_train, y_train_multi[cols])
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertTrue(hasattr(model.pipeline, "classes_"))
        remove_dir(model_dir)


        ############
        # Test continue training
        tfidf = TfidfVectorizer()
        svc = LinearSVC()
        pipeline = Pipeline([('tfidf', tfidf), ('svc', svc)])
        model = ModelPipeline(model_dir=model_dir, pipeline=pipeline, multi_label=False)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        self.assertFalse(hasattr(model.pipeline['svc'], "classes_"))
        model.fit(x_train, y_train_mono)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertTrue(hasattr(model.pipeline['svc'], "classes_"))
        # Second fit
        with self.assertRaises(RuntimeError):
            model.fit(x_train[:50], y_train_mono[:50])
        self.assertEqual(model_dir, model.model_dir)
        remove_dir(model_dir)

    def test03_model_pipeline_predict(self):
        '''Test of the method predict of {{package_name}}.models_training.model_pipeline.ModelPipeline'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        x_train = np.array(["ceci est un test", "pas cela", "cela non plus", "ici test", "là, rien!"])
        y_train_mono = np.array([0, 1, 0, 1, 0])
        y_train_multi = pd.DataFrame({'test1': [0, 0, 0, 1, 0], 'test2': [1, 0, 0, 0, 0], 'test3': [0, 0, 0, 1, 0]})
        cols = ['test1', 'test2', 'test3']

        # Mono-label
        tfidf = TfidfVectorizer()
        rf = RandomForestClassifier(n_estimators=10)
        pipeline = Pipeline([('tfidf', tfidf), ('rf', rf)])
        model = ModelPipeline(model_dir=model_dir, pipeline=pipeline, multi_label=False)
        model.fit(x_train, y_train_mono)
        preds = model.predict(x_train, return_proba=False)
        self.assertEqual(preds.shape, (len(x_train),))
        preds = model.predict('test', return_proba=False)
        self.assertEqual(preds, model.predict(['test'], return_proba=False)[0])
        proba = model.predict(x_train, return_proba=True)
        self.assertEqual(proba.shape, (len(x_train), 2))
        proba = model.predict('test', return_proba=True)
        self.assertEqual([elem for elem in proba], [elem for elem in model.predict(['test'], return_proba=True)[0]])
        remove_dir(model_dir)

        # Multi-labels
        tfidf = TfidfVectorizer()
        rf = RandomForestClassifier(n_estimators=10)
        pipeline = Pipeline([('tfidf', tfidf), ('rf', OneVsRestClassifier(rf))])
        model = ModelPipeline(model_dir=model_dir, pipeline=pipeline, multi_label=True)
        model.fit(x_train, y_train_multi[cols])
        preds = model.predict(x_train)
        self.assertEqual(preds.shape, (len(x_train), len(cols)))
        preds = model.predict('test', return_proba=False)
        self.assertEqual([elem for elem in preds], [elem for elem in model.predict(['test'], return_proba=False)[0]])
        proba = model.predict(x_train, return_proba=True)
        self.assertEqual(proba.shape, (len(x_train), len(cols)))
        proba = model.predict('test', return_proba=True)
        self.assertEqual([elem for elem in proba], [elem for elem in model.predict(['test'], return_proba=True)[0]])
        remove_dir(model_dir)

        # Model needs to be fitted
        with self.assertRaises(AttributeError):
            model = ModelPipeline(model_dir=model_dir, pipeline=pipeline, multi_label=False)
            model.predict('test')
        remove_dir(model_dir)

    def test04_model_pipeline_predict_proba(self):
        '''Test of the method predict_proba of {{package_name}}.models_training.model_pipeline.ModelPipeline'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        x_train = np.array(["ceci est un test", "pas cela", "cela non plus", "ici test", "là, rien!"])
        y_train_mono = np.array([0, 1, 0, 1, 0])
        y_train_multi = pd.DataFrame({'test1': [0, 0, 0, 1, 0], 'test2': [1, 0, 0, 0, 0], 'test3': [0, 0, 0, 1, 0]})
        cols = ['test1', 'test2', 'test3']

        # Mono-label
        tfidf = TfidfVectorizer()
        rf = RandomForestClassifier(n_estimators=10)
        pipeline = Pipeline([('tfidf', tfidf), ('rf', rf)])
        model = ModelPipeline(model_dir=model_dir, pipeline=pipeline, multi_label=False)
        model.fit(x_train, y_train_mono)
        preds = model.predict_proba(x_train)
        self.assertEqual(preds.shape, (len(x_train), 2))
        preds = model.predict_proba('test')
        self.assertEqual([elem for elem in preds], [elem for elem in model.predict_proba(['test'])[0]])
        remove_dir(model_dir)

        # Multi-labels
        tfidf = TfidfVectorizer()
        rf = RandomForestClassifier(n_estimators=10)
        pipeline = Pipeline([('tfidf', tfidf), ('rf', OneVsRestClassifier(rf))])
        model = ModelPipeline(model_dir=model_dir, pipeline=pipeline, multi_label=True)
        model.fit(x_train, y_train_multi[cols])
        preds = model.predict_proba(x_train)
        self.assertEqual(preds.shape, (len(x_train), len(cols)))
        preds = model.predict_proba('test')
        self.assertEqual([elem for elem in preds], [elem for elem in model.predict_proba(['test'])[0]])
        remove_dir(model_dir)

        # Model needs to be fitted
        with self.assertRaises(AttributeError):
            model = ModelPipeline(model_dir=model_dir, pipeline=pipeline, multi_label=False)
            model.predict_proba('test')
        remove_dir(model_dir)

    def test05_model_pipeline_save(self):
        '''Test of the method save of {{package_name}}.models_training.model_pipeline.ModelPipeline'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # With Pipeline
        tfidf = TfidfVectorizer()
        rf = RandomForestClassifier(n_estimators=10)
        pipeline = Pipeline([('tfidf', tfidf), ('rf', rf)])
        model = ModelPipeline(model_dir=model_dir, pipeline=pipeline)
        model.save(json_data={'test': 8})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"{model.model_name}.pkl")))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"sklearn_pipeline_standalone.pkl")))
        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertTrue('package_version' in configs.keys())
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertTrue('model_name' in configs.keys())
        self.assertTrue('model_dir' in configs.keys())
        self.assertTrue('trained' in configs.keys())
        self.assertTrue('nb_fit' in configs.keys())
        self.assertTrue('list_classes' in configs.keys())
        self.assertTrue('dict_classes' in configs.keys())
        self.assertTrue('x_col' in configs.keys())
        self.assertTrue('y_col' in configs.keys())
        self.assertTrue('multi_label' in configs.keys())
        self.assertTrue('level_save' in configs.keys())
        self.assertTrue('librairie' in configs.keys())
        self.assertEqual(configs['librairie'], 'scikit-learn')
        # Specific model used
        self.assertTrue('tfidf_confs' in configs.keys())
        self.assertTrue('rf_confs' in configs.keys())
        remove_dir(model_dir)

        # Without Pipeline
        model = ModelPipeline(model_dir=model_dir, pipeline=None)
        model.save(json_data={'test': 8})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"{model.model_name}.pkl")))
        self.assertFalse(os.path.exists(os.path.join(model.model_dir, f"sklearn_pipeline_standalone.pkl")))
        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertTrue('package_version' in configs.keys())
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertTrue('model_name' in configs.keys())
        self.assertTrue('model_dir' in configs.keys())
        self.assertTrue('trained' in configs.keys())
        self.assertTrue('nb_fit' in configs.keys())
        self.assertTrue('list_classes' in configs.keys())
        self.assertTrue('dict_classes' in configs.keys())
        self.assertTrue('x_col' in configs.keys())
        self.assertTrue('y_col' in configs.keys())
        self.assertTrue('multi_label' in configs.keys())
        self.assertTrue('level_save' in configs.keys())
        self.assertTrue('librairie' in configs.keys())
        self.assertEqual(configs['librairie'], 'scikit-learn')
        # Specific model used
        self.assertTrue('tfidf_confs' not in configs.keys())
        self.assertTrue('rf_confs' not in configs.keys())
        remove_dir(model_dir)

        # WITH level_save = 'LOW'
        tfidf = TfidfVectorizer()
        rf = RandomForestClassifier(n_estimators=10)
        pipeline = Pipeline([('tfidf', tfidf), ('rf', rf)])
        model = ModelPipeline(model_dir=model_dir, pipeline=pipeline, level_save='LOW')
        model.save(json_data={'test': 8})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertFalse(os.path.exists(os.path.join(model.model_dir, f"{model.model_name}.pkl")))
        self.assertFalse(os.path.exists(os.path.join(model.model_dir, f"sklearn_pipeline_standalone.pkl")))
        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertTrue('package_version' in configs.keys())
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertTrue('model_name' in configs.keys())
        self.assertTrue('model_dir' in configs.keys())
        self.assertTrue('trained' in configs.keys())
        self.assertTrue('nb_fit' in configs.keys())
        self.assertTrue('list_classes' in configs.keys())
        self.assertTrue('dict_classes' in configs.keys())
        self.assertTrue('x_col' in configs.keys())
        self.assertTrue('y_col' in configs.keys())
        self.assertTrue('multi_label' in configs.keys())
        self.assertTrue('level_save' in configs.keys())
        self.assertTrue('librairie' in configs.keys())
        self.assertEqual(configs['librairie'], 'scikit-learn')
        # Specific model used
        self.assertTrue('tfidf_confs' in configs.keys())
        self.assertTrue('rf_confs' in configs.keys())
        remove_dir(model_dir)


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3

## Generic rules model for text classification
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Classes :
# - ModelRules -> Generic rules model


# Get logger
import re
import logging
import numpy as np
import pandas as pd
from functools import partial
from typing import Union, Tuple, List, Any

from {{package_name}} import utils
from {{package_name}}.models_training.model_class import ModelClass


logger = logging.getLogger(__name__)


class ModelRules(ModelClass):
    '''Generic rules model for text classification
    The rules (contained in the table 'rules') each have three attributes:
    name (facultative, given on the fly if not explicit), pattern (a regex)
    and cls, the returned class if the pattern is found.
    '''

    _default_name = 'model_rules'

    def __init__(self, default: Union[str, None] = None, **kwargs) -> None:
        '''Initialization of the class (see ModelClass for more arguments)

        Kwargs:
            default (str) : The default prediction if no match
        '''
        self.default = default
        self.rules: List[dict] = []
        super().__init__(**kwargs)

    def create_rule(self, name: Union[str, None] = None, pattern: str="*", cls: str="classe_1") -> None:
        '''Creates a rule and appends it to the set of rules.

        Kwargs:
            name (str) : The name of the rule
            pattern (str) : The regex used to match
            cls (str) : The class to give if the pattern matches

        '''
        self.rules.append({'name': 'rule_'+str(len(self.rules)) if name is None else name,
                           'pattern': pattern,
                           'cls': cls})

    def validate_rule(self, rule: dict) -> Tuple[bool, Union[dict, None]]:
        '''Validates if a rule is ok or not : attributes present and with the correct type. Creates a name on the fly if needed.

        Args:
            rule (dict) : The rule to validate
        '''
        k = rule.keys()
        ret = True
        if 'name' in k:
            if not isinstance(rule['name'], str):
                logger.warning('For a rule, \'name\' must be of type str, automatic renaming')
                rule['name'] = 'rule_' + str(len(self.rules))
        else:
            rule['name'] = 'rule_' + str(len(self.rules))

        if 'pattern' not in k or not isinstance(rule['pattern'], str):
            logger.warning('For a rule, \'pattern\' must be of type str and present')
            ret = False

        if 'cls' not in k or not isinstance(type(rule['cls']), str):
            logger.warning('For a rule, \'cls\' must be of type str and present')
            ret = False
        if ret is False:
            rule = None

        return ret, rule

    def add_rules(self, rules: Union[List[dict], dict]) -> None:
        ''' Adds a set of rules (list) or a pre-formated rule (dict) to the rules already present

        Args:
            rules (list<dict> or dict) : the rules to add
        '''
        if isinstance(rules, dict):
            rules = [rules]

        for rule in rules:
            ok, rule = self.validate_rule(rule)
            if ok:
                self.rules.append(rule)

    def fit(self, x_train, y_train, x_valid=None, y_valid=None, **kwargs) -> None:
        '''"Trains" the model by compiling the regex'''
        # If not multi-labels, transform y_train as dummies (should already be the case for multi-labels)
        if not self.multi_label:
            # If len(array.shape)==2, we flatten the array if the second dimension is useless
            if isinstance(y_train, np.ndarray) and len(y_train.shape) == 2 and y_train.shape[1] == 1:
                y_train = np.ravel(y_train)
            if isinstance(y_valid, np.ndarray) and len(y_valid.shape) == 2 and y_valid.shape[1] == 1:
                y_valid = np.ravel(y_valid)
            # Transformation dummies
            y_train_dummies = pd.get_dummies(y_train)
            y_valid_dummies = pd.get_dummies(y_valid) if y_valid is not None else None
            self.list_classes = list(y_train_dummies.columns)
        # Else keep it as it
        else:
            y_train_dummies = y_train
            y_valid_dummies = y_valid
            if hasattr(y_train_dummies, 'columns'):
                self.list_classes = list(y_train_dummies.columns)
            else:
                logger.warning("Can't read the name of the columns of y_train -> inverse transformation won't be possible")
                # We still create a list of classes in order to be compatible with other functions
                self.list_classes = [str(_) for _ in range(pd.DataFrame(y_train_dummies).shape[1])]

        # Set dict_classes based on list classes
        self.dict_classes = {i: col for i, col in enumerate(self.list_classes)}
        self.index_classes = {col: i for i, col in enumerate(self.list_classes)}
        for rule in self.rules:
            rule['regexp'] = re.compile(rule['pattern'])

        # Set trained
        self.trained = True
        self.nb_fit += 1

    def apply_rules(self, text: str, cls=None) -> Union[int, str, None]:
        '''Applies the set of rules on a text. Gives the label of the class if cls is not None otherwise
        gives 1 (used in the multi-labels case)

        Args:
            text (str) : The text on which to apply the rules
            cls (str) : If not None, the function gives the label of the class
        Return:
            The self.default, the label of the class, 0 or 1
        '''
        match = None
        for rule in self.rules:
            if rule["regexp"].search(text):
                if cls is not None:
                    match = rule["cls"]
                    break
                else:
                    match = 1 if rule["cls"] is None else 0

        if match is None:
            match = self.default
        return match

    @utils.data_agnostic_str_to_list
    def predict(self, x_test: Union[List[str], pd.Series]) -> Any:
        ''''Predicts' on the test dataset by applying the set of rules, either all at the same time either by class
        in the multi-labels/multi-classes (TODO : the distinction does not exist for now)

        Args:
            x_test (list<str> or pd.series) : The list of text to apply the rules to
        Return:
            (?) : The predictions
        '''
        x_test = pd.Series(x_test)
        if self.multi_label:
            preds = pd.DataFrame()
            for i in self.list_classes:
                preds[i] = partial(self.apply_rules, cls=i)
        else:
            preds = x_test.apply(self.apply_rules)

        return preds


if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3

## Model TF-IDF + Dense
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Classes :
# - ModelTfidfDense -> Model for predictions via TF-IDF + Dense

import os
import json
import pickle
import shutil
import logging
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from typing import Union, Any, List, Callable
from sklearn.feature_extraction.text import TfidfVectorizer

import tensorflow as tf
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.models import Model, Sequential
from tensorflow.keras.models import load_model as load_model_keras
from tensorflow.keras.layers import (ELU, LSTM, AveragePooling1D,
                                            BatchNormalization, Bidirectional,
                                            Conv1D, Dense, Dropout, Embedding,
                                            Flatten, GlobalAveragePooling1D,
                                            GlobalMaxPooling1D, Input,
                                            LeakyReLU, MaxPooling1D, ReLU,
                                            SpatialDropout1D, concatenate)

from {{package_name}} import utils
from {{package_name}}.models_training.model_keras import ModelKeras
from {{package_name}}.models_training import utils_deep_keras, utils_models


sns.set(style="darkgrid")


class ModelTfidfDense(ModelKeras):
    '''Model for predictions via TF-IDF + Dense'''

    _default_name = 'model_tfidf_dense'

    def __init__(self, tfidf_params: Union[dict, None] = None, **kwargs) -> None:
        '''Initialization of the class (see ModelClass & ModelKeras for more arguments).

        Kwargs:
            tfidf_params (dict) : Parameters for the tfidf
        '''
        # Init.
        super().__init__(**kwargs)

        # Get logger (must be done after super init)
        self.logger = logging.getLogger(__name__)

        if tfidf_params is None:
            tfidf_params = {}
        self.tfidf = TfidfVectorizer(**tfidf_params)

    @utils.data_agnostic_str_to_list
    @utils.trained_needed
    def predict_proba(self, x_test, experimental_version: bool = False, **kwargs) -> np.ndarray:
        '''Probabilies predictions on the test dataset.

        Args:
            x_test (?): Array-like or sparse matrix, shape = [n_samples]
        Kwargs:
            experimental_version (bool): If an experimental (but faster) version must be used
        Returns:
            (np.ndarray): Array, shape = [n_samples, n_classes]
        '''
        # Prepare input
        x_test = self._prepare_x_test(x_test)
        # Process
        if experimental_version:
            return self.experimental_predict_proba(x_test)
        else:
            return self.model.predict(x_test, batch_size=128, verbose=1)  # type: ignore

    def _prepare_x_train(self, x_train) -> np.ndarray:
        '''Prepares the input data for the model. Called when fitting the model

        Args:
            x_train (?): Array-like, shape = [n_samples, n_features]
        Returns:
            (np.ndarray): Prepared data
        '''
        # Fit tfidf & return x transformed
        self.tfidf.fit(x_train)
        # TODO: Use of todense because tensorflow 2.3 does not support sparse data anymore
        return self.tfidf.transform(x_train).todense()

    def _prepare_x_test(self, x_test) -> np.ndarray:
        '''Prepares the input data for the model. Called when fitting the model

        Args:
            x_train (?): Array-like, shape = [n_samples, n_features]
        Returns:
            (np.ndarray): Prepared data
        '''
        # Get tf-idf & fit on train
        # TODO: Use of todense because tensorflow 2.3 does not support sparse data anymore
        return self.tfidf.transform(x_test).todense()

    def _get_model(self) -> Model:
        '''Gets a model structure

        Returns:
            (Model): a Keras model
        '''
        # Get input/output dimensions
        input_dim = len(self.tfidf.get_feature_names())
        num_classes = len(self.list_classes)

        # Process
        model = Sequential()

        model.add(Dense(128, activation=None, kernel_initializer='he_uniform', input_shape=(input_dim,)))
        model.add(BatchNormalization(momentum=0.9))
        model.add(ELU(alpha=1.0))
        model.add(Dropout(0.5))

        model.add(Dense(64, activation=None, kernel_initializer='he_uniform'))
        model.add(BatchNormalization(momentum=0.9))
        model.add(ELU(alpha=1.0))
        model.add(Dropout(0.5))

        model.add(Dense(32, activation=None, kernel_initializer='he_uniform'))
        model.add(BatchNormalization(momentum=0.9))
        model.add(ELU(alpha=1.0))
        model.add(Dropout(0.5))

        # Last layer
        activation = 'sigmoid' if self.multi_label else 'softmax'
        model.add(Dense(num_classes, activation=activation, kernel_initializer='glorot_uniform'))

        # Compile model
        lr = self.keras_params.get('learning_rate', 0.002)
        decay = self.keras_params.get('decay', 0.0)
        self.logger.info(f"Learning rate utilisée : {lr}")
        self.logger.info(f"Decay utilisé : {decay}")
        optimizer = Adam(lr=lr, decay=decay)
        # loss = utils_deep_keras.f1_loss if self.multi_label else 'categorical_crossentropy'
        loss = 'binary_crossentropy' if self.multi_label else 'categorical_crossentropy'  # utils_deep_keras.f1_loss also possible if multi-labels
        metrics: List[Union[str, Callable]] = ['accuracy'] if not self.multi_label else ['categorical_accuracy', utils_deep_keras.f1, utils_deep_keras.precision, utils_deep_keras.recall]
        model.compile(optimizer=optimizer, loss=loss, metrics=metrics)
        if self.logger.getEffectiveLevel() < logging.ERROR:
            model.summary()

        # Try to save model as png if level_save > 'LOW'
        if self.level_save in ['MEDIUM', 'HIGH']:
            self._save_model_png(model)

        # Return
        return model

    def save(self, json_data: Union[dict, None] = None) -> None:
        '''Saves the model

        Kwargs:
            json_data (dict): Additional configurations to be saved
        '''
        # Save configuration JSON
        if json_data is None:
            json_data = {}

        # Add tfidf params
        confs = self.tfidf.get_params()
        # Get rid of some non serializable conf
        for special_conf in ['dtype', 'base_estimator']:
            if special_conf in confs.keys():
                confs[special_conf] = str(confs[special_conf])
        json_data['tfidf_confs'] = confs

        # Save tfidf if not None & level_save > LOW
        if (self.tfidf is not None) and (self.level_save in ['MEDIUM', 'HIGH']):
            # Manage paths
            tfidf_path = os.path.join(self.model_dir, f"tfidf_standalone.pkl")
            # Save as pickle
            with open(tfidf_path, 'wb') as f:
                # TODO: use dill to get rid of  "can't pickle ..." errors
                pickle.dump(self.tfidf, f)

        # Save
        super().save(json_data=json_data)

    def reload_from_standalone(self, **kwargs) -> None:
        '''Reloads a model from its configuration and "standalones" files
        - /!\\ Experimental /!\\ -

        Kwargs:
            configuration_path (str): Path to configuration file
            hdf5_path (str): Path to hdf5 file
            tf_idf_path (str): Path to tfidf file
        Raises:
            ValueError: If configuration_path is None
            ValueError: If hdf5_path is None
            ValueError: If tf_idf_path is None
            FileNotFoundError: If the object configuration_path is not an existing file
            FileNotFoundError: If the object hdf5_path is not an existing file
            FileNotFoundError: If the object tf_idf_path is not an existing file
        '''
        # Retrieve args
        configuration_path = kwargs.get('configuration_path', None)
        hdf5_path = kwargs.get('hdf5_path', None)
        tfidf_path = kwargs.get('tfidf_path', None)

        # Checks
        if configuration_path is None:
            raise ValueError("The argument configuration_path can't be None")
        if hdf5_path is None:
            raise ValueError("The argument hdf5_path can't be None")
        if tfidf_path is None:
            raise ValueError("The argument tfidf_path can't be None")
        if not os.path.exists(configuration_path):
            raise FileNotFoundError(f"The file {configuration_path} does not exist")
        if not os.path.exists(hdf5_path):
            raise FileNotFoundError(f"The file {hdf5_path} does not exist")
        if not os.path.exists(tfidf_path):
            raise FileNotFoundError(f"The file {tfidf_path} does not exist")

        # Load confs
        with open(configuration_path, 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        # Can't set int as keys in json, so need to cast it after reloading
        # dict_classes keys are always ints
        if 'dict_classes' in configs.keys():
            configs['dict_classes'] = {int(k): v for k, v in configs['dict_classes'].items()}
        elif 'list_classes' in configs.keys():
            configs['dict_classes'] = {i: col for i, col in enumerate(configs['list_classes'])}

        # Set class vars
        # self.model_name = # Keep the created name
        # self.model_dir = # Keep the created folder
        self.nb_fit = configs.get('nb_fit', 1)  # Consider one unique fit by default
        self.trained = configs.get('trained', True)  # Consider trained by default
        # Try to read the following attributes from configs and, if absent, keep the current one
        for attribute in ['x_col', 'y_col',
                          'list_classes', 'dict_classes', 'multi_label', 'level_save',
                          'batch_size', 'epochs', 'validation_split', 'patience',
                          'nb_iter_keras', 'keras_params', 'embedding_name']:
            setattr(self, attribute, configs.get(attribute, getattr(self, attribute)))

        # Reload model
        self.model = load_model_keras(hdf5_path, custom_objects=self.custom_objects)

        # Save best hdf5 in new folder
        new_hdf5_path = os.path.join(self.model_dir, 'best.hdf5')
        shutil.copyfile(hdf5_path, new_hdf5_path)

        # Reload tfidf
        with open(tfidf_path, 'rb') as f:
            self.tfidf = pickle.load(f)


if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import pickle
import shutil
import numpy as np
import pandas as pd

from {{package_name}} import utils
from {{package_name}}.models_training.model_tfidf_svm import ModelTfidfSvm
from {{package_name}}.monitoring.model_explainer import LimeExplainer, AttentionExplainer
from {{package_name}}.models_training.model_embedding_lstm_structured_attention import ModelEmbeddingLstmStructuredAttention


# Disable logging
import logging
logging.disable(logging.CRITICAL)


def remove_dir(path):
    if os.path.isdir(path): shutil.rmtree(path)


class ModelExplainerTest(unittest.TestCase):
    '''Main class to test model_explainer'''

    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)
        # Check if data folder exists
        data_path = utils.get_data_path()
        if not os.path.exists(data_path):
            os.mkdir(data_path)
        # Create a mock embedding
        fake_embedding = {'toto': [0.25, 0.90, 0.12], 'titi': [0.85, 0.12, 0.8], 'test': [0.5, 0.6, 0.1]}
        fake_path = os.path.join(data_path, 'fake_embedding.pkl')
        if os.path.exists(fake_path):
            os.remove(fake_path)
        with open(fake_path, 'wb') as f:
            pickle.dump(fake_embedding, f, pickle.HIGHEST_PROTOCOL)

    def test01_lime_explainer_nominal(self):
        '''Test of the mono-class mono-label case'''

        # Model creation
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model = ModelTfidfSvm(model_dir=model_dir, multi_label=False)
        # fake model_conf
        model_conf = {'preprocess_str': 'no_preprocess'}
        # Set vars
        x_train = np.array(["ceci est un test", "pas cela", "cela non plus", "ici test", "là, rien!"] * 100)
        x_test = np.array(["cela est un test", "ni cela", "non plus", "ici test", "là, rien de rien!"] * 100)
        y_train_mono = np.array([0, 1, 0, 1, 2] * 100)
        y_test_mono = y_train_mono.copy()
        y_train_multi = pd.DataFrame({'test1': [0, 0, 0, 1, 0] * 100, 'test2': [1, 0, 0, 0, 0] * 100, 'test3': [0, 0, 0, 1, 0] * 100})
        y_test_multi = y_train_multi.copy()
        cols = ['test1', 'test2', 'test3']

        # Mono-label
        model = ModelTfidfSvm(model_dir=model_dir, multi_label=False)
        model.fit(x_train, y_train_mono)
        exp = LimeExplainer(model, model_conf)
        txt = exp.explain_instance_as_list(text="ceci est un test", classes=[0, 1])
        html = exp.explain_instance_as_html(text="ceci est un test", classes=[0, 1])
        remove_dir(model_dir)

        # Multi-labels
        model = ModelTfidfSvm(model_dir=model_dir, multi_label=True)
        model.fit(x_train, y_train_multi[cols])
        txt = exp.explain_instance_as_list(text="ceci est un test", classes=[0, 1])
        html = exp.explain_instance_as_html(text="ceci est un test", classes=[0, 1])
        remove_dir(model_dir)

    def test02_attention_explainer_nominal(self):
        '''Test of the mono-class mono-label case'''

        # Model creation
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        x_train = np.array(["ceci est un test", "pas cela", "cela non plus", "ici test", "là, rien!"] * 100)
        x_test = np.array(["cela est un test", "ni cela", "non plus", "ici test", "là, rien de rien!"] * 100)
        y_train_mono = np.array([0, 1, 0, 1, 2] * 100)
        y_test_mono = y_train_mono.copy()
        y_train_multi = pd.DataFrame({'test1': [0, 0, 0, 1, 0] * 100, 'test2': [1, 0, 0, 0, 0] * 100, 'test3': [0, 0, 0, 1, 0] * 100})
        y_test_multi = y_train_multi.copy()
        cols = ['test1', 'test2', 'test3']

        # Mono-label
        model = ModelEmbeddingLstmStructuredAttention(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False,
                                            max_sequence_length=10, max_words=100,
                                            padding='pre', truncating='post',
                                            embedding_name='fake_embedding.pkl')
        model.fit(x_train, y_train_mono)
        exp = AttentionExplainer(model)
        txt = exp.explain_instance_as_list(text="ceci est un test")
        html = exp.explain_instance_as_html(text="ceci est un test")
        remove_dir(model_dir)


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3

# Apply a Machine Learning algorithm to obtain predictions
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Ex: python 3_predict.py  --filename dataset_test.csv --x_col preprocessed_text --model_dir model_tfidf_svm_2021_04_15-10_23_13 --y_col Survived


import os
import json
import logging
import argparse
import numpy as np
import pandas as pd
from pathlib import Path
from datetime import datetime
from typing import Union, List, Tuple

from {{package_name}} import utils
from {{package_name}}.preprocessing import preprocess
from {{package_name}}.models_training import utils_models

# Get logger
logger = logging.getLogger('{{package_name}}.3_predict')


def main(filename: str, x_col: Union[str, int], model_dir: str, y_col: Union[List[Union[str, int]], None] = None,
         sep: str = '{{default_sep}}', encoding: str = '{{default_encoding}}') -> None:
    '''Gets a model's predictions on a given dataset

    Args:
        filename (str): Name of the test dataset (actually a path relative to {{package_name}}-data)
            It must not be preprocessed, hence not have metadata in its first line
        x_col (str | int): Name of the model's input column - x
        model_dir (str): Name of the model to use (not a path, just the directory name)
    Kwargs:
        y_col (list<str|int>): Name of the model's target column(s) - y
            Only used to get some metrics if available
        sep (str): Separator to use with the .csv files
        encoding (str): Encoding to use with the .csv files
    Raises:
        ValueError: If multi-labels and bad OHE format
    '''

    ##############################################
    # Loading model & data
    ##############################################

    # Load model
    logger.info("Loading model ...")
    model, model_conf = utils_models.load_model(model_dir=model_dir)

    # Load dataset & preprocess it
    logger.info("Loading and preprocessing test dataset ...")
    df, preprocess_col = load_dataset_test(filename, x_col, model_conf, sep=sep, encoding=encoding)


    ##############################################
    # Get predictions
    ##############################################

    # Get predictions
    logger.info("Getting predictions ...")
    y_pred = model.predict(df[preprocess_col], return_proba=False)
    predictions_col = utils.get_new_column_name(list(df.columns), 'predictions')
    df[predictions_col] = list(model.inverse_transform(np.array(y_pred)))


    ##############################################
    # Save results
    ##############################################

    # Remove preprocessing from dataset (we do not need it anymore)
    df = df[[col for col in df.columns if col != preprocess_col]]

    # Save result
    logger.info("Saving results ...")
    save_dir = os.path.join(utils.get_data_path(), 'predictions', Path(filename).stem, datetime.now().strftime("predictions_%Y_%m_%d-%H_%M_%S"))
    if not os.path.isdir(save_dir):
        os.makedirs(save_dir)
    save_file = "predictions.csv"
    file_path = os.path.join(save_dir, save_file)
    df.to_csv(file_path, sep='{{default_sep}}', encoding='{{default_encoding}}', index=None)

    # Also save some info into a configs file
    conf_file = 'configurations.json'
    conf_path = os.path.join(save_dir, conf_file)
    conf = {
        'model_dir': model_dir,
        'preprocess_str': model_conf['preprocess_str'],
        'model_name': model_conf['model_name']
    }
    with open(conf_path, 'w', encoding='{{default_encoding}}') as f:
        json.dump(conf, f, indent=4)


    ##############################################
    # Get metrics
    ##############################################

    # Get metrics if y_col is not None
    if y_col is not None:

        ### INFO: target data must have a correct format (cf training file)
        if len(y_col) > 1:
            try:
                y_true = df[y_col].astype(int)  # Need to cast OHE var into integers
                for col in y_col:
                    assert sorted(y_true[y_col].unique()) == [0, 1]
            except:
                raise ValueError("You provided several target columns, but at least one of them does not seem to be in a correct OHE format.")
        else:
            y_true = df[y_col[0]]
            # No need to cast target in string, already done by the data loader

        cols_to_add: List[pd.Series] = []  # You can add columns to save here
        series_to_add = [df[col] for col in cols_to_add]
        # Change model directory to save dir & get preds
        model.model_dir = save_dir
        model.get_and_save_metrics(y_true, y_pred, series_to_add=series_to_add, type_data='with_y_true')


def load_dataset_test(filename: str, x_col: Union[str, int], model_conf: dict,
                      sep: str = '{{default_sep}}', encoding: str = '{{default_encoding}}') -> Tuple[pd.DataFrame, str]:
    '''Function to load a test dataset & preprocess it

    Args:
        filename (str): Name of the dataset to load (actually a path relative to {{package_name}}-data)
        x_col (str | int): Name of the model's input column - x
        model_conf (dict): Model configuration
    Kwargs:
        sep (str): Separator to use with the .csv files
        encoding (str): Encoding to use with the .csv files
    Raises:
        FileNotFoundError: If the file does not exist in {{package_name}}-data
    Returns:
        pd.DataFrame: Loaded & preprocessed dataframe
        str: Column name of the preprocessed data
    '''
    logger.info(f"Loading a dataset ({filename})")

    # Get dataset
    data_path = utils.get_data_path()
    file_path = os.path.join(data_path, filename)
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"The file {file_path} does not exist")

    # Get dataset
    df = pd.read_csv(file_path, sep=sep, encoding=encoding, dtype=str).fillna('')

    # Get preprocessor from model conf
    preprocess_str = model_conf['preprocess_str']
    preprocessor = preprocess.get_preprocessor(preprocess_str)

    # Apply preprocessing
    preprocess_col = utils.get_new_column_name(list(df.columns), 'preprocessed_column')
    # We do not check for column presence in dataframe
    # TODO: add check ?
    logger.info("Preprocessing test data")
    df[preprocess_col] = preprocessor(df[x_col])

    # Return
    return df, preprocess_col


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--filename', default='newdata.csv', help="Name of the test dataset (actually a path relative to {{package_name}}-data)")
    parser.add_argument('-x', '--x_col', required=True, help="Name of the model's input column - x")
    # model_X should be the model's directory name: e.g. model_tfidf_svm_2019_12_05-12_57_18
    parser.add_argument('-m', '--model_dir', required=True, help="Name of the model to use (not a path, just the directory name)")
    parser.add_argument('-y', '--y_col', nargs='+', default=None, help="Name of the model's target column(s) - y")
    parser.add_argument('--sep', default='{{default_sep}}', help="Separator to use with the .csv files.")
    parser.add_argument('--encoding', default="{{default_encoding}}", help="Encoding to use with the .csv files.")
    parser.add_argument('--force_cpu', dest='on_cpu', action='store_true', help="Whether to force training on CPU (and not GPU)")
    parser.set_defaults(on_cpu=False)
    args = parser.parse_args()
    # Check forced CPU usage
    if args.on_cpu:
        os.environ['CUDA_VISIBLE_DEVICES'] = ""
        logger.info("----------------------------")
        logger.info("CPU USAGE FORCED BY THE USER")
        logger.info("----------------------------")
    # Main
    main(filename=args.filename, x_col=args.x_col, model_dir=args.model_dir,
         y_col=args.y_col, sep=args.sep, encoding=args.encoding)
